from typing import Union
import numpy as np
import scipy.linalg
from casadi import SX, vertcat, sin, cos, Function, jacobian
from acados_template import (
    AcadosOcp,
    AcadosOcpSolver,
    AcadosModel,
    AcadosOcpConstraints,
)


ACADOS_MULTIPLIER_ORDER = [
    "lbu",
    "lbx",
    "lg",
    "lh",
    "lphi",
    "ubu",
    "ubx",
    "ug",
    "uh",
    "uphi",
    "lsbu",
    "lsbx",
    "lsg",
    "lsh",
    "lsphi",
    "usbu",
    "usbx",
    "usg",
    "ush",
    "usphi",
]


def rename_key_in_dict(d: dict, old_key: str, new_key: str):
    d[new_key] = d.pop(old_key)

    return d


def rename_item_in_list(lst: list, old_item: str, new_item: str):
    if old_item in lst:
        index_old = lst.index(old_item)
        lst[index_old] = new_item

    return lst


class LagrangeMultiplierExtractor(object):
    """
    Class to store dimensions of constraints
    """

    order: list = ACADOS_MULTIPLIER_ORDER

    idx_at_stage: list

    def __init__(self, constraints: AcadosOcpConstraints, N: int = 20):
        super().__init__()

        replacements = {
            0: [("lbx", "lbx_0"), ("ubx", "ubx_0")],
            N: [
                ("lbx", "lbx_e"),
                ("ubx", "ubx_e"),
                ("lg", "lg_e"),
                ("ug", "ug_e"),
                ("lh", "lh_e"),
                ("uh", "uh_e"),
                ("lphi", "lphi_e"),
                ("uphi", "uphi_e"),
                ("lsbx", "lsbx_e"),
                ("usbx", "usbx_e"),
                ("lsg", "lsg_e"),
                ("usg", "usg_e"),
                ("lsh", "lsh_e"),
                ("ush", "ush_e"),
                ("lsphi", "lsphi_e"),
                ("usphi", "usphi_e"),
            ],
        }

        idx_at_stage = [dict.fromkeys(self.order, 0) for _ in range(N + 1)]

        for stage, keys in replacements.items():
            for old_key, new_key in keys:
                idx_at_stage[stage] = rename_key_in_dict(idx_at_stage[stage], old_key, new_key)

        # Loop over all constraints and count the number of constraints of each type. Store the indices in a dict.
        for stage, idx in enumerate(idx_at_stage):
            _start = 0
            _end = 0
            for attr in dir(constraints):
                if idx.keys().__contains__(attr):
                    _end += len(getattr(constraints, attr))
                    idx[attr] = slice(_start, _end)
                    _start = _end

        self.idx_at_stage = idx_at_stage

    def get_idx_at_stage(self, stage: int, field: str) -> slice:
        """
        Get the indices of the constraints of the given type at the given stage.

        Parameters:
            stage: stage index
            field: constraint type

        Returns:
            indices: slice object
        """
        return self.idx_at_stage[stage][field]

    def __call__(self, stage: int, field: str, lam: np.ndarray) -> np.ndarray:
        """
        Extract the multipliers at the given stage from the vector of multipliers.

        Parameters:
            stage: stage index
            field: constraint type
            lam: vector of multipliers

        Returns:
            lam: vector of multipliers at the given stage and of the given type
        """
        return lam[self.get_idx_at_stage(stage, field)]


def build_cost_functions(
    ocp: AcadosOcp,
) -> tuple[Function, Function, Function, Function]:
    """
    Build the cost functions for the OCP.

    Parameters:
        acados_ocp: acados OCP object

    Returns:
        fun_l: stage cost function
        fun_m: terminal cost function
        fun_dl_dp: derivative of stage cost function with respect to parameters
        fun_dm_dp: derivative of terminal cost function with respect to parameters
    """

    W = ocp.cost.W
    W_e = ocp.cost.W_e

    yref = ocp.cost.yref
    yref_e = ocp.cost.yref_e

    x = ocp.model.x
    u = ocp.model.u
    p = ocp.model.p

    if ocp.cost.cost_type == "LINEAR_LS":
        """
        In case of LINEAR_LS:
        stage cost is
        :math:`l(x,u,z) = || V_x \, x + V_u \, u + V_z \, z - y_\\text{ref}||^2_W`,
        terminal cost is
        :math:`m(x) = || V^e_x \, x - y_\\text{ref}^e||^2_{W^e}`
        """

        cost_y_expr = ocp.cost.Vx @ ocp.model.x + ocp.cost.Vu @ ocp.model.u
        cost_y_expr_e = ocp.cost.Vx_e @ ocp.model.x

    elif ocp.cost.cost_type == "NONLINEAR_LS":
        cost_y_expr = ocp.model.cost_y_expr
        cost_y_expr_e = ocp.model.cost_y_expr_e

    else:
        raise NotImplementedError("Only LINEAR_LS and NONLINEAR_LS cost types are supported at the moment.")

    l = (cost_y_expr - yref).T @ W @ (cost_y_expr - yref)
    m = (cost_y_expr_e - yref_e).T @ W_e @ (cost_y_expr_e - yref_e)

    fun_l = Function("l", [x, u, p], [l], ["x", "u", "p"], ["l"])
    fun_m = Function("m", [x, p], [m], ["x", "p"], ["m"])
    fun_dl_dp = Function("dl_dp", [x, u, p], [jacobian(l, p)], ["x", "u", "p"], ["dl_dp"])
    fun_dm_dp = Function("dm_dp", [x, p], [jacobian(m, p)], ["x", "p"], ["dm_dp"])

    return fun_l, fun_m, fun_dl_dp, fun_dm_dp


def build_discrete_dynamics_functions(ocp: AcadosOcp) -> tuple[Function, Function]:
    """
    Build the discrete dynamics functions for the OCP.

    Parameters:
        acados_ocp: acados OCP object

    Returns:
        fun_f: discrete dynamics function
        fun_df_dp: derivative of discrete dynamics function with respect to parameters
    """

    x = ocp.model.x
    u = ocp.model.u
    p = ocp.model.p
    f = ocp.model.disc_dyn_expr

    # Build the discrete dynamics function
    if ocp.solver_options.integrator_type == "ERK" and ocp.solver_options.sim_method_num_stages == 4:
        f = ERK4(f, x, u, p)
    elif ocp.solver_options.integrator_type == "DISCRETE":
        f = ocp.model.disc_dyn_expr
    else:
        raise NotImplementedError("Only ERK4 and DISCRETE integrator types are supported at the moment.")

    fun_f = Function("f", [x, u, p], [f], ["x", "u", "p"], ["xf"])
    fun_df_dp = Function("df_dp", [x, u, p], [jacobian(f, p)], ["x", "u", "p"], ["dxf_dp"])

    return fun_f, fun_df_dp


def build_constraint_functions(
    ocp: AcadosOcp,
) -> tuple[Function, Function, Function, Function]:
    """
    Build the constraint functions for the OCP.

    Parameters:
        acados_ocp: acados OCP object

    Returns:
        fun_g: stage constraint function
        fun_dg_dp: derivative of stage constraint function with respect to parameters
        fun_g_e: terminal constraint function
        fun_dg_dp_e: derivative of terminal constraint function with respect to parameters
    """
    pass


class LagrangeFunction(object):
    """
    Lagrange function for the OCP
    """

    # f_theta: Function
    # ocp: AcadosOcp

    lam_extractor: LagrangeMultiplierExtractor
    fun_f: Function  # Discrete dynamics
    fun_df_dp: Function  # Derivative of discrete dynamics with respect to parameters
    fun_l: Function  # Stage cost
    fun_m: Function  # Terminal cost
    fun_dl_dp: Function  # Derivative of stage cost with respect to parameters
    fun_dm_dp: Function  # Derivative of terminal cost with respect to parameters

    def __init__(self, acados_ocp_solver: AcadosOcpSolver):
        super().__init__()

        self.lam_extractor = LagrangeMultiplierExtractor(
            acados_ocp_solver.acados_ocp.constraints,
            acados_ocp_solver.acados_ocp.dims.N,
        )

        self.fun_f, self.fun_df_dp = build_discrete_dynamics_functions(acados_ocp_solver.acados_ocp)
        self.fun_l, self.fun_m, self.fun_dl_dp, self.fun_dm_dp = build_cost_functions(acados_ocp_solver.acados_ocp)
        _ = build_constraint_functions(acados_ocp_solver.acados_ocp)

    def __call__(self, acados_ocp_solver: AcadosOcpSolver, p: np.ndarray) -> np.float32:
        """
        Evaluate the Lagrange function at the current solution of the OCP.
        """

        res = 0.0

        if True:
            constraints = acados_ocp_solver.acados_ocp.constraints

            # Initial condition equality constraint
            stage = 0
            res += self.lam_extractor(stage, "lbx_0", acados_ocp_solver.get(stage, "lam")) @ (
                getattr(constraints, "lbx_0") - acados_ocp_solver.get(stage, "x")
            )

            res += self.lam_extractor(stage, "ubx_0", acados_ocp_solver.get(stage, "lam")) @ (
                getattr(constraints, "ubx_0") - acados_ocp_solver.get(stage, "x")
            )

            # Inequality constraints at stage k
            if getattr(constraints, "lbx").size > 0:
                for stage in range(1, acados_ocp_solver.acados_ocp.dims.N):
                    res += self.lam_extractor(stage, "lbx", acados_ocp_solver.get(stage, "lam")) @ (
                        getattr(constraints, "lbx") - acados_ocp_solver.get(stage, "x")
                    )

            if getattr(constraints, "ubx").size > 0:
                for stage in range(1, acados_ocp_solver.acados_ocp.dims.N):
                    res += self.lam_extractor(stage, "ubx", acados_ocp_solver.get(stage, "lam")) @ (
                        getattr(constraints, "ubx") - acados_ocp_solver.get(stage, "x")
                    )

            if getattr(constraints, "lbu").size > 0:
                for stage in range(1, acados_ocp_solver.acados_ocp.dims.N):
                    res += self.lam_extractor(stage, "lbu", acados_ocp_solver.get(stage, "lam")) @ (
                        getattr(constraints, "lbu") - acados_ocp_solver.get(stage, "u")
                    )

            if getattr(constraints, "ubu").size > 0:
                for stage in range(1, acados_ocp_solver.acados_ocp.dims.N):
                    res += self.lam_extractor(stage, "ubu", acados_ocp_solver.get(stage, "lam")) @ (
                        getattr(constraints, "ubu") - acados_ocp_solver.get(stage, "u")
                    )

            # Terminal state inequality constraints
            stage = acados_ocp_solver.acados_ocp.dims.N
            if getattr(constraints, "lbx_e").size > 0:
                res += self.lam_extractor(stage, "lbx_e", acados_ocp_solver.get(stage, "lam")) @ (
                    getattr(constraints, "lbx_e") - acados_ocp_solver.get(stage, "x")
                )

            if getattr(constraints, "ubx_e").size > 0:
                res += self.lam_extractor(stage, "ubx_e", acados_ocp_solver.get(stage, "lam")) @ (
                    getattr(constraints, "ubx_e") - acados_ocp_solver.get(stage, "x")
                )

        res += acados_ocp_solver.get_cost()

        # Dynamic equality constraint
        res_dyn = 0
        for stage in range(acados_ocp_solver.acados_ocp.dims.N - 1):
            x = acados_ocp_solver.get(stage, "x")
            xnext = acados_ocp_solver.get(stage + 1, "x")
            u = acados_ocp_solver.get(stage, "u")
            pi = acados_ocp_solver.get(stage, "pi")
            res_dyn += pi @ (self.eval_f(x=x, u=u, p=p) - xnext)

        res += res_dyn

        return res

    def eval_f(self, x: np.ndarray, u: np.ndarray, p: np.ndarray) -> np.ndarray:
        """
        Evaluate the integrator at the current solution of the OCP.

        Parameters:
            x: state
            u: control
            p: parameters

        Returns:
            xf: integrated state
        """
        return self.fun_f(x=x, u=u, p=p)["xf"].full().flatten()

    def eval_df_dp(self, x: np.ndarray, u: np.ndarray, p: np.ndarray) -> np.ndarray:
        """
        Evaluate the gradient of the integrator with respect to the parameters at the current solution of the OCP.

        Parameters:
            x: state
            u: control
            p: parameters

        Returns:
            df_dp: gradient of the integrator with respect to the parameters
        """
        return self.fun_df_dp(x=x, u=u, p=p)["dxf_dp"].full().flatten()

    def eval_dL_dp(self, acados_ocp_solver: AcadosOcpSolver, p: np.ndarray) -> np.ndarray:
        """
        Evaluate the gradient of the Lagrange function with respect to the parameters at the current solution of the OCP.

        Parameters:
            acados_ocp_solver: acados OCP solver object with the current solution
            p: parameters

        Returns:
            dL_dp: gradient of the Lagrange function with respect to the parameters
        """

        res = 0
        # Dynamic equality constraint
        for stage in range(acados_ocp_solver.acados_ocp.dims.N - 1):
            x = acados_ocp_solver.get(stage, "x")
            u = acados_ocp_solver.get(stage, "u")
            pi = acados_ocp_solver.get(stage, "pi")
            res += pi @ self.eval_df_dp(x=x, u=u, p=p)

        return res


def ERK4(
    f: Union[SX, Function],
    x: Union[SX, np.ndarray],
    u: Union[SX, np.ndarray],
    p: Union[SX, np.ndarray],
    h: float,
) -> Union[SX, np.ndarray]:
    """
    Explicit Runge-Kutta 4 integrator

    TODO: Works for numeric values as well as for symbolic values. Type hinting is a bit misleading.

    Parameters:
        f: function to integrate
        x: state
        u: control
        p: parameters
        h: step size

        Returns:
            xf: integrated state
    """
    k1 = f(x, u, p)
    k2 = f(x + h / 2 * k1, u, p)
    k3 = f(x + h / 2 * k2, u, p)
    k4 = f(x + h * k3, u, p)
    xf = x + h / 6 * (k1 + 2 * k2 + 2 * k3 + k4)

    return xf


def export_acados_ocp() -> AcadosOcp:
    """
    Define the acados OCP solver object.

    Parameters:
        ocp: acados OCP object

    Returns:
        acados_ocp_solver: acados OCP solver object
    """

    # create ocp object to formulate the OCP
    ocp = AcadosOcp()

    # set model
    model = export_pendulum_ode_model()

    Tf = 1.0
    N = 20
    dT = Tf / N

    ode = Function("ode", [model.x, model.u, model.p], [model.f_expl_expr])
    model.disc_dyn_expr = ERK4(ode, model.x, model.u, model.p, dT)
    print("built ERK4 for pendulum model with dT = ", dT)

    ###
    ocp.model = model

    nx = model.x.size()[0]
    nu = model.u.size()[0]
    ny = nx + nu
    ny_e = nx

    # set dimensions
    ocp.dims.N = N

    # set cost module
    ocp.cost.cost_type = "LINEAR_LS"
    ocp.cost.cost_type_e = "LINEAR_LS"

    Q = 2 * np.diag([1e3, 1e3, 1e-2, 1e-2])
    R = 2 * np.diag([1e-1])

    ocp.cost.W = scipy.linalg.block_diag(Q, R)

    ocp.cost.W_e = Q

    ocp.cost.Vx = np.zeros((ny, nx))
    ocp.cost.Vx[:nx, :nx] = np.eye(nx)

    Vu = np.zeros((ny, nu))
    Vu[4, 0] = 1.0
    ocp.cost.Vu = Vu

    ocp.cost.Vx_e = np.eye(nx)

    ocp.cost.yref = np.array((0.0, np.pi, 0.0, 0.0, 0.0))
    ocp.cost.yref_e = np.array((0.0, np.pi, 0.0, 0.0))

    # set constraints
    Fmax = 80
    x0 = np.array([0.5, 0.0, 0.0, 0.0])
    ocp.constraints.lbu = np.array([-Fmax])
    ocp.constraints.ubu = np.array([+Fmax])
    ocp.constraints.x0 = x0
    ocp.constraints.idxbu = np.array([0])

    # ocp.constraints.lbx = np.array([-2.0, -2 * np.pi, -10.0, -10.0])
    # ocp.constraints.ubx = np.array([2.0, 2 * np.pi, 10.0, 10.0])
    # ocp.constraints.idxbx = np.array([0, 1, 2, 3])

    ocp.solver_options.qp_solver = "PARTIAL_CONDENSING_HPIPM"  # FULL_CONDENSING_QPOASES
    ocp.solver_options.hessian_approx = "GAUSS_NEWTON"
    # ocp.solver_options.integrator_type = "DISCRETE"
    ocp.solver_options.integrator_type = "DISCRETE"
    ocp.solver_options.nlp_solver_type = "SQP"  # SQP_RTI

    ocp.solver_options.qp_solver_cond_N = N

    ocp.solver_options.qp_solver_iter_max = 200

    # set prediction horizon
    ocp.solver_options.tf = Tf

    ocp.parameter_values = np.array([1.0])

    return ocp


def export_q_value_acados_ocp() -> AcadosOcp:
    """
    Define the acados OCP solver object.

    Parameters:
        ocp: acados OCP object

    Returns:
        acados_ocp_solver: acados OCP solver object
    """

    # create ocp object to formulate the OCP
    ocp = AcadosOcp()

    # set model
    model = export_pendulum_ode_model()

    Tf = 1.0
    N = 20
    dT = Tf / N

    ode = Function("ode", [model.x, model.u, model.p], [model.f_expl_expr])
    model.disc_dyn_expr = ERK4(ode, model.x, model.u, model.p, dT)
    print("built ERK4 for pendulum model with dT = ", dT)

    ###
    ocp.model = model

    nx = model.x.size()[0]
    nu = model.u.size()[0]
    ny = nx + nu
    ny_e = nx

    # set dimensions
    ocp.dims.N = N

    # set cost module
    ocp.cost.cost_type = "LINEAR_LS"
    ocp.cost.cost_type_e = "LINEAR_LS"

    Q = 2 * np.diag([1e3, 1e3, 1e-2, 1e-2])
    R = 2 * np.diag([1e-1])

    ocp.cost.W = scipy.linalg.block_diag(Q, R)

    ocp.cost.W_e = Q

    ocp.cost.Vx = np.zeros((ny, nx))
    ocp.cost.Vx[:nx, :nx] = np.eye(nx)

    Vu = np.zeros((ny, nu))
    Vu[4, 0] = 1.0
    ocp.cost.Vu = Vu

    ocp.cost.Vx_e = np.eye(nx)

    ocp.cost.yref = np.array((0.0, 0.0, 0.0, 0.0, 0.0))
    ocp.cost.yref_e = np.array((0.0, 0.0, 0.0, 0.0))

    # set constraints
    Fmax = 80
    x0 = np.array([0.5, 0.0, 0.0, 0.0])
    ocp.constraints.lbu = np.array([-Fmax])
    ocp.constraints.ubu = np.array([+Fmax])
    ocp.constraints.x0 = x0
    ocp.constraints.idxbu = np.array([0])
    ocp.constraints.idxsbu = np.array([0])

    ocp.cost.zu = np.array([1e3])
    ocp.cost.zl = np.array([1e3])
    ocp.cost.Zl = np.array([1e3])
    ocp.cost.Zu = np.array([1e3])

    # ocp.constraints.lbx = np.array([-2.0, -2 * np.pi, -10.0, -10.0])
    # ocp.constraints.ubx = np.array([2.0, 2 * np.pi, 10.0, 10.0])
    # ocp.constraints.idxbx = np.array([0, 1, 2, 3])

    ocp.solver_options.qp_solver = "PARTIAL_CONDENSING_HPIPM"  # FULL_CONDENSING_QPOASES
    ocp.solver_options.hessian_approx = "GAUSS_NEWTON"
    # ocp.solver_options.integrator_type = "DISCRETE"
    ocp.solver_options.integrator_type = "DISCRETE"
    ocp.solver_options.nlp_solver_type = "SQP"  # SQP_RTI

    ocp.solver_options.qp_solver_cond_N = N

    ocp.solver_options.qp_solver_iter_max = 200

    # set prediction horizon
    ocp.solver_options.tf = Tf

    ocp.parameter_values = np.array([1.0])

    return ocp


def export_pendulum_ode_model() -> AcadosModel:
    model_name = "pendulum"

    # constants
    if False:
        m_cart = 1.0  # mass of the cart [kg]
    else:
        m_cart = SX.sym("m_cart")

    m = 0.1  # mass of the ball [kg]
    g = 9.81  # gravity constant [m/s^2]
    l = 0.8  # length of the rod [m]

    # set up states & controls
    x1 = SX.sym("x1")
    theta = SX.sym("theta")
    v1 = SX.sym("v1")
    dtheta = SX.sym("dtheta")

    x = vertcat(x1, theta, v1, dtheta)

    F = SX.sym("F")
    u = vertcat(F)

    # xdot
    x1_dot = SX.sym("x1_dot")
    theta_dot = SX.sym("theta_dot")
    v1_dot = SX.sym("v1_dot")
    dtheta_dot = SX.sym("dtheta_dot")

    xdot = vertcat(x1_dot, theta_dot, v1_dot, dtheta_dot)

    # algebraic variables
    # z = None

    # parameters
    p = m_cart

    # dynamics
    cos_theta = cos(theta)
    sin_theta = sin(theta)
    denominator = m_cart + m - m * cos_theta * cos_theta
    f_expl = vertcat(
        v1,
        dtheta,
        (-m * l * sin_theta * dtheta * dtheta + m * g * cos_theta * sin_theta + F) / denominator,
        (-m * l * cos_theta * sin_theta * dtheta * dtheta + F * cos_theta + (m_cart + m) * g * sin_theta)
        / (l * denominator),
    )

    f_impl = xdot - f_expl

    model = AcadosModel()

    model.f_impl_expr = f_impl
    model.f_expl_expr = f_expl
    model.x = x
    model.xdot = xdot
    model.u = u
    # model.z = z
    model.p = p
    model.name = model_name

    return model


def update_parameters(acados_ocp_solver: AcadosOcpSolver, p: np.ndarray) -> AcadosOcpSolver:
    """
    Update the parameters of the OCP solver.

    Parameters:
        acados_ocp_solver: acados OCP solver object
        p: parameters
    """
    for stage in range(acados_ocp_solver.acados_ocp.dims.N):
        acados_ocp_solver.set(stage, "p", p)

    return acados_ocp_solver
